#!/usr/bin/env python
# encoding: utf-8
#
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2023 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2023 Florian Spreckelsen <f.spreckelsen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.
#
"""Module for converting a data model into a json schema compatible dictionary.

The scope of this json schema is the automatic generation of user interfaces.
"""

from collections import OrderedDict
from typing import Any, Dict, Iterable, List, Optional, Union

import linkahead as db
from linkahead.common.datatype import get_list_datatype, is_list_datatype


class JsonSchemaExporter:
    """A class which collects everything needed for the conversion.
    """

    def __init__(self, additional_properties: bool = True,
                 name_property_for_new_records: bool = False,
                 description_property_for_new_records: bool = False,
                 additional_options_for_text_props: dict = None,
                 units_in_description: bool = True,
                 do_not_create: List[str] = None,
                 do_not_retrieve: List[str] = None,
                 no_remote: bool = False,
                 ):
        """Set up a JsonSchemaExporter, which can then be applied on RecordTypes.

        Parameters
        ----------
        additional_properties : bool, optional
            Whether additional properties will be admitted in the resulting
            schema. Optional, default is True.
        name_property_for_new_records : bool, optional
            Whether objects shall generally have a `name` property in the generated schema.
            Optional, default is False.
        description_property_for_new_records : bool, optional
            Whether objects shall generally have a `description` property in the generated schema.
            Optional, default is False.
        additional_options_for_text_props : dict, optional
            Dictionary containing additional "pattern" or "format" options for
            string-typed properties. Optional, default is empty.
        units_in_description : bool, optional
            Whether to add the unit of a LinkAhead property (if it has any) to the
            description of the corresponding schema entry. If set to false, an
            additional `unit` key is added to the schema itself which is purely
            annotational and ignored, e.g., in validation. Default is True.
        do_not_create : list[str]
            A list of RedcordType names, for which there should be no option
            to create them.  Instead, only the choice of existing elements should
            be given.
        do_not_retrieve : list[str]
            A list of RedcordType names, for which no Records shall be retrieved.  Instead, only an
            object description should be given.  If this list overlaps with the `do_not_create`
            parameter, the behavior is undefined.
        no_remote : bool
            If True, do not attempt to connect to a LinkAhead server at all. Default is False.
        """
        if not additional_options_for_text_props:
            additional_options_for_text_props = {}
        if not do_not_create:
            do_not_create = []
        if not do_not_retrieve:
            do_not_retrieve = []

        self._additional_properties = additional_properties
        self._name_property_for_new_records = name_property_for_new_records
        self._description_property_for_new_records = description_property_for_new_records
        self._additional_options_for_text_props = additional_options_for_text_props
        self._units_in_description = units_in_description
        self._do_not_create = do_not_create
        self._do_not_retrieve = do_not_retrieve
        self._no_remote = no_remote

    @staticmethod
    def _make_required_list(rt: db.RecordType):
        """Return the list of names of properties with importance db.OBLIGATORY."""
        required_list = []
        for prop in rt.properties:
            if rt.get_importance(prop.name) != db.OBLIGATORY:
                continue
            prop_name = prop.name
            if isinstance(prop.datatype, db.Entity):
                prop_name = prop.datatype.name
            required_list.append(prop_name)

        return required_list

    def _make_segment_from_prop(self, prop: db.Property):
        """Return the JSON Schema segment for the given property

        Parameters
        ----------
        prop : db.Property
            The property to be transformed.
        """

        if prop.datatype == db.TEXT or prop.datatype == db.DATETIME:
            text_format = None
            text_pattern = None
            if prop.name in self._additional_options_for_text_props:
                if "pattern" in self._additional_options_for_text_props[prop.name]:
                    text_pattern = self._additional_options_for_text_props[prop.name]["pattern"]
                if "format" in self._additional_options_for_text_props[prop.name]:
                    text_format = self._additional_options_for_text_props[prop.name]["format"]
                elif prop.datatype == db.DATETIME:
                    # Set the date or datetime format if only a pattern is given ...
                    text_format = ["date", "date-time"]
            elif prop.datatype == db.DATETIME:
                # ... again, for those props that don't appear in the additional
                # options list.
                text_format = ["date", "date-time"]

            return self._make_text_property(prop.description, text_format, text_pattern)

        json_prop = OrderedDict()
        if prop.description:
            json_prop["description"] = prop.description
        if self._units_in_description and prop.unit:
            if "description" in json_prop:
                json_prop["description"] += f" Unit is {prop.unit}."
            else:
                json_prop["description"] = f"Unit is {prop.unit}."
        elif prop.unit:
            json_prop["unit"] = prop.unit

        if prop.datatype == db.BOOLEAN:
            json_prop["type"] = "boolean"
        elif prop.datatype == db.INTEGER:
            json_prop["type"] = "integer"
        elif prop.datatype == db.DOUBLE:
            json_prop["type"] = "number"
        elif is_list_datatype(prop.datatype):
            json_prop["type"] = "array"
            list_element_prop = db.Property(
                name=prop.name, datatype=get_list_datatype(prop.datatype, strict=True))
            json_prop["items"] = self._make_segment_from_prop(list_element_prop)
        elif prop.is_reference():
            if prop.datatype == db.REFERENCE:
                # No Record creation since no RT is specified and we don't know what
                # schema to use, so only enum of all Records and all Files.
                values = self._retrieve_enum_values("RECORD") + self._retrieve_enum_values("FILE")
                json_prop["enum"] = values
            elif prop.datatype == db.FILE:
                json_prop["type"] = "string"
                json_prop["format"] = "data-url"
            else:
                prop_name = prop.datatype
                if isinstance(prop.datatype, db.Entity):
                    prop_name = prop.datatype.name
                if prop_name in self._do_not_retrieve:
                    values = []
                else:
                    values = self._retrieve_enum_values(f"RECORD '{prop_name}'")
                if prop_name in self._do_not_create:
                    # Only a simple list of values
                    json_prop["enum"] = values
                else:
                    if self._no_remote:
                        rt = prop.datatype
                    else:
                        rt = db.execute_query(f"FIND RECORDTYPE WITH name='{prop_name}'",
                                              unique=True)
                    subschema = self._make_segment_from_recordtype(rt)
                    if values:
                        subschema["title"] = "Create new"
                        json_prop["oneOf"] = [
                            {
                                "title": "Existing entries",
                                "enum": values,
                            },
                            subschema
                        ]
                    else:
                        json_prop = subschema

        else:
            raise ValueError(
                f"Unknown or no property datatype. Property {prop.name} with type {prop.datatype}")

        return json_prop

    @staticmethod
    def _make_text_property(description="", text_format=None, text_pattern=None):
        prop = {
            "type": "string"
        }
        if description:
            prop["description"] = description
        if text_format is not None:
            if isinstance(text_format, list):
                # We want the type inside the options, not in the head:
                # "datetime property": {
                #   "anyOf": [
                #     {
                #       "type": "string",
                #       "format": "date"
                #     },
                #     {
                #       "type": "string",
                #       "format": "date-time"
                #     }]}
                prop.pop("type")
                prop["anyOf"] = [{"type": "string", "format": tf} for tf in text_format]
            else:
                prop["format"] = text_format
        if text_pattern is not None:
            prop["pattern"] = text_pattern

        return prop

    def _retrieve_enum_values(self, role: str):

        if self._no_remote:
            return []

        possible_values = db.execute_query(f"SELECT name, id FROM {role}")

        vals = []
        for val in possible_values:
            if val.name:
                vals.append(f"{val.id}, {val.name}")
            else:
                vals.append(f"{val.id}")

        return vals

    def _make_segment_from_recordtype(self, rt: db.RecordType):
        """Return a Json schema segment for the given RecordType.
        """
        schema: dict[str, Any] = {
            "type": "object"
        }

        schema["required"] = self._make_required_list(rt)
        schema["additionalProperties"] = self._additional_properties

        props = OrderedDict()
        if self._name_property_for_new_records:
            props["name"] = self._make_text_property("The name of the Record to be created")
        if self._description_property_for_new_records:
            props["description"] = self._make_text_property(
                "The description of the Record to be created")

        for prop in rt.properties:
            if prop.name in props:
                # Multi property
                raise NotImplementedError(
                    "Creating a schema for multi-properties is not specified. "
                    f"Property {prop.name} occurs more than once."
                )
            props[prop.name] = self._make_segment_from_prop(prop)

        schema["properties"] = props

        return schema

    def recordtype_to_json_schema(self, rt: db.RecordType):
        """Create a jsonschema from a given RecordType that can be used, e.g., to
        validate a json specifying a record of the given type.

        Parameters
        ----------
        rt : RecordType
            The RecordType from which a json schema will be created.

        Returns
        -------
        schema : dict
            A dict containing the json schema created from the given RecordType's properties.
        """
        if rt is None:
            raise ValueError(
                "recordtype_to_json_schema(...) cannot be called with a `None` RecordType.")
        schema = self._make_segment_from_recordtype(rt)
        schema["$schema"] = "https://json-schema.org/draft/2019-09/schema"
        if rt.name:
            schema["title"] = rt.name
        if rt.description:
            schema["description"] = rt.description

        return schema


def recordtype_to_json_schema(rt: db.RecordType, additional_properties: bool = True,
                              name_property_for_new_records: bool = False,
                              description_property_for_new_records: bool = False,
                              additional_options_for_text_props: Optional[dict] = None,
                              units_in_description: bool = True,
                              do_not_create: List[str] = None,
                              do_not_retrieve: List[str] = None,
                              no_remote: bool = False,
                              ):
    """Create a jsonschema from a given RecordType that can be used, e.g., to
    validate a json specifying a record of the given type.

    This is a standalone function which works without manually creating a
    JsonSchemaExporter object.

    Parameters
    ----------
    rt : RecordType
        The RecordType from which a json schema will be created.
    additional_properties : bool, optional
        Whether additional properties will be admitted in the resulting
        schema. Optional, default is True.
    name_property_for_new_records : bool, optional
        Whether objects shall generally have a `name` property in the generated schema. Optional,
        default is False.
    description_property_for_new_records : bool, optional
        Whether objects shall generally have a `description` property in the generated schema.
        Optional, default is False.
    additional_options_for_text_props : dict, optional
        Dictionary containing additional "pattern" or "format" options for
        string-typed properties. Optional, default is empty.
    units_in_description : bool, optional
        Whether to add the unit of a LinkAhead property (if it has any) to the
        description of the corresponding schema entry. If set to false, an
        additional `unit` key is added to the schema itself which is purely
        annotational and ignored, e.g., in validation. Default is True.
    do_not_create : list[str]
        A list of RedcordType names, for which there should be no option
        to create them.  Instead, only the choice of existing elements should
        be given.
    do_not_retrieve : list[str]
        A list of RedcordType names, for which no Records shall be retrieved.  Instead, only an
        object description should be given.  If this list overlaps with the `do_not_create`
        parameter, the behavior is undefined.
    no_remote : bool
        If True, do not attempt to connect to a LinkAhead server at all.  Default is False.

    Returns
    -------
    schema : dict
        A dict containing the json schema created from the given RecordType's properties.

    """

    exporter = JsonSchemaExporter(
        additional_properties=additional_properties,
        name_property_for_new_records=name_property_for_new_records,
        description_property_for_new_records=description_property_for_new_records,
        additional_options_for_text_props=additional_options_for_text_props,
        units_in_description=units_in_description,
        do_not_create=do_not_create,
        do_not_retrieve=do_not_retrieve,
        no_remote=no_remote,
    )
    return exporter.recordtype_to_json_schema(rt)


def make_array(schema: dict) -> dict:
    """Create an array of the given schema.

The result will look like this:

.. code:: js

  { "type": "array",
    "items": {
        // the schema
      }
  }

Parameters
----------

schema : dict
  The JSON schema which shall be packed into an array.

Returns
-------

out : dict
  A JSON schema dict with a top-level array which contains instances of the given schema.
    """
    result = {
        "type": "array",
        "items": schema,
        "$schema": "https://json-schema.org/draft/2019-09/schema",
    }
    return result


def merge_schemas(schemas: Union[Dict[str, dict], Iterable[dict]]) -> dict:
    """Merge the given schemata into a single schema.

The result will look like this:

.. code:: js

  {
    "type": "object",
    "properties": {
      // A, B, C
    },
    "required": [
      // "A", "B", "C"
    ],
    "additionalProperties": false
  }


Parameters
----------

schemas : dict[str, dict] | Iterable[dict]
  A dict or iterable of schemata which shall be merged together.  If this is a dict, the keys will
  be used as property names, otherwise the titles of the submitted schemata.  If they have no title,
  numbers will be used as a fallback.  Note that even with a dict, the original schema's "title" is
  not changed.

Returns
-------

out : dict
  A JSON schema dict with a top-level object which contains the given schemata as properties.
    """
    sub_schemas: dict[str, dict] = OrderedDict()
    required = []

    if isinstance(schemas, dict):
        sub_schemas = schemas
        required = [str(k) for k in schemas.keys()]
    else:
        for i, schema in enumerate(schemas, start=1):
            title = schema.get("title", str(i))
            sub_schemas[title] = schema
            required.append(title)

    result = {
        "type": "object",
        "properties": sub_schemas,
        "required": required,
        "additionalProperties": False,
        "$schema": "https://json-schema.org/draft/2020-12/schema",
    }

    return result
